|
| Glue routines between C programs and the Motorola 68040 FPSP library routines
|
| Eric Norum                     		eric@skatter.usask.ca
| Saskatchewan Accelerator Laboratory
| University of Saskatchewan
| Saskatoon, Canada.             		NeXTMail accepted.

.include "l_fpsp.h"

|
| A bit of trickery to keep the loader from pulling in NeXT's
| shared-library versions of these routines.
|
	.globl .file_definition_libsys_s.B.shlib_trig.o
	.set .file_definition_libsys_s.B.shlib_trig.o,0
	.globl .file_definition_libsys_s.B.shlib_asincos.o
	.set .file_definition_libsys_s.B.shlib_asincos.o,0
	.globl .file_definition_libsys_s.B.shlib_atan.o
	.set .file_definition_libsys_s.B.shlib_atan.o,0
	.globl .file_definition_libsys_s.B.shlib_atan2.o
	.set .file_definition_libsys_s.B.shlib_atan2.o,0

	.globl .file_definition_libsys_s.B.shlib_sinh.o
	.set .file_definition_libsys_s.B.shlib_sinh.o,0
	.globl .file_definition_libsys_s.B.shlib_cosh.o
	.set .file_definition_libsys_s.B.shlib_cosh.o,0
	.globl .file_definition_libsys_s.B.shlib_tanh.o
	.set .file_definition_libsys_s.B.shlib_tanh.o,0
	.globl .file_definition_libsys_s.B.shlib_asinh.o
	.set .file_definition_libsys_s.B.shlib_asinh.o,0
	.globl .file_definition_libsys_s.B.shlib_acosh.o
	.set .file_definition_libsys_s.B.shlib_acosh.o,0
	.globl .file_definition_libsys_s.B.shlib_atanh.o
	.set .file_definition_libsys_s.B.shlib_atanh.o,0

	.globl .file_definition_libsys_s.B.shlib_exp.o
	.set .file_definition_libsys_s.B.shlib_exp.o,0
	.globl .file_definition_libsys_s.B.shlib_expm1.o
	.set .file_definition_libsys_s.B.shlib_expm1.o,0
	.globl .file_definition_libsys_s.B.shlib_log.o
	.set .file_definition_libsys_s.B.shlib_log.o,0
	.globl .file_definition_libsys_s.B.shlib_log10.o
	.set .file_definition_libsys_s.B.shlib_log10.o,0
	.globl .file_definition_libsys_s.B.shlib_log1p.o
	.set .file_definition_libsys_s.B.shlib_log1p.o,0
	.globl .file_definition_libsys_s.B.shlib_pow.o
	.set .file_definition_libsys_s.B.shlib_pow.o,0

|
| The easy ones.  They map directly to M68040FPSP routines.
| None of these do anything with `errno', though.
| A small speed gain could be made by just renaming the appropriate
| M68040FPSP entry point.  I have chosen to leave the FPSP routines
| just as they came from Motorola.
|
	.globl	_sin	; _sin:		; jmp	fsind
	.globl	_cos	; _cos:		; jmp	fcosd
	.globl	_tan	; _tan:		; jmp	ftand
	.globl	_asin	; _asin:	; jmp	fasind
	.globl	_acos	; _acos:	; jmp	facosd
	.globl	_atan	; _atan:	; jmp	fatand
	.globl	_exp	; _exp:		; jmp	fetoxd
	.globl	_expm1	; _expm1:	; jmp	fetoxm1d
	.globl	_log	; _log:		; jmp	flognd
	.globl	_log10	; _log10:	; jmp	flog10d
	.globl	_log1p	; _log1p:	; jmp	flognp1d
	.globl	_sinh	; _sinh:	; jmp	fsinhd
	.globl	_cosh	; _cosh:	; jmp	fcoshd
	.globl	_tanh	; _tanh:	; jmp	ftanhd
	.globl	_atanh	; _atanh:	; jmp	fatanhd

|
| Routines that aren't in math.h, but that have M68040FPSP counterparts
|
	.globl _flog2	; _flog2:	; jmp	flog2d
	.globl _fint	; _fint:	; jmp	fintd
	.globl _fintrz	; _fintrz:	; jmp	fintrzd

|
| The Motorola-supplied entry (fsincosd in l_entry.s) is wrong
| so we have to do it ourselves.
|
	.globl	_sincos
_sincos:
	link	a6,#-LOCAL_SIZE
	fmovemx fp2-fp3,a6@(USER_FP2)
	fmovel	fpsr,a6@(USER_FPSR)
	fmovel	fpcr,a6@(USER_FPCR)
	fmovel	fpcr,d1		| user's rounding mode/precision
	fmovel	#0,fpcr		| force rounding mode/prec to extended,rn
|
|	copy, convert and tag input argument
|
	fmoved	a6@(8),fp0
	fmovex	fp0,a6@(ETEMP)
	lea	a6@(ETEMP),a0
	jbsr	tag
	moveb	d0,a6@(STAG)
	tstb	d0
	jne	L_1117
	jbsr	ssincos		| normalized (regular) number
	jra	L_111B
L_1117:
	cmpb	#0x20,d0		| zero?
	jne	L_1118
	jbsr	szero
	fmoved	#0e1.0,fp1
	jra	L_111B
L_1118:
	cmpb	#0x40,d0		| infinity?
	jne	L_1119
	jbsr	t_operr
	fmovex	fp0,fp1
	jra	L_111B
L_1119:
	cmpb	#0x60,d0		| NaN?
	jne	L_111A
	jbsr	mon_nan
	fmovex	fp0,fp1
	jra	L_111B
L_111A:
	jbsr	ssincosd		| assuming a denorm...

L_111B:
	fmovel	fpsr,d0		| update status register
	orb	a6@(USER_FPSR+3),d0	|add previously accrued exceptions
	fmovel	d0,fpsr
|
|	Result is now in FP0/FP1
|
	fmovemx a6@(USER_FP2),fp2-fp3	| note: FP0 not restored
	movl	a6@(16),a0		| get pointer to sin result
	fmoved	fp0,a0@			| save sin result
	movl	a6@(20),a0		| get pointer to cos result
	fmoved	fp1,a0@			| save cos result
	unlk	a6
	rts

|
| Some simple-minded attempts at functions that
| don't map directly to M68040FPSP routines.
| `Simple-minded' because they don't take into account
| all the possibilities of NaN's, Infinities, Denormalized
| numbers, etc.
|
	.globl _atan2
_atan2:
	fmoved	sp@(12),fp0	| fetch X argument
	fjeq	xz		| it's zero, skip
	fjgt	xp		| it's positive, skip
xm:
	fmoved	sp@(4),fp1	| fetch Y argument
	fjeq	xmyz		| it's zero , skip
	fjgt	xmyp		| it's positive, return pi/2
xmym:
	fdivx	fp0,fp1		| form y/x
	fmoved	fp1,sp@-	| push y/x
	jbsr	fatand		| compute atan(y/x)
	addl	#8,sp		| pop argument
	fsubx	PPI,fp0		| convert to quadrant 3
	bras	atan2done	| return value
xmyp:
	fdivx	fp0,fp1		| form y/x
	fmovex	fp1,sp@-	| push y/x
	jbsr	fatanx		| compute atan(y/x)
	addl	#12,sp		| pop argument
	faddx	PPI,fp0		| convert to quadrant 2
	bras	atan2done	| return value
xmyz:
	fmovex	PPI,fp0		| return PI
	bras	atan2done
xp:
	fmoved	sp@(4),fp1	| fetch Y argument
	fjeq	xpyz		| it's zero , skip
xpyp:
xpym:
	fdivx	fp0,fp1		| form y/x
	fmoved	fp1,sp@-	| push y/x
	jbsr	fatand		| compute atan(y/x)
	addl	#8,sp		| pop argument
	bras	atan2done	| return value
xpyz:
	fmovex	fp1,fp0		| return 0.0
	bras	atan2done
xz:
	fmoved	sp@(4),fp1	| fetch Y argument
	fjeq	xzyz		| it's zero too, skip
	fjgt	xzyp		| it's positive, return pi/2
xzym:
	fmovex	NPIBY2,fp0	| it's negative, return -pi/2
	bras	atan2done	| return the value
xzyp:
	fmovex	PPIBY2,fp0	| it's negative, return -pi/2
	bras	atan2done	| return the value
xzyz:
	fmoves	NANS,fp0	| return NAN
atan2done:
	fmoved	fp0,sp@-	| Archaic NeXT compiler returns doubles in d0/d1
	movel	sp@+,d0
	movel	sp@+,d1
	rts
PPI:	.long 0x40000000,0xC90FDAA2,0x2168C235
PPIBY2:	.long 0x3FFF0000,0xC90FDAA2,0x2168C235
NPIBY2:	.long 0xBFFF0000,0xC90FDAA2,0x2168C235

	.globl _pow
_pow:
	fmoved	sp@(12),fp0	| fetch Y argument
	fjne	pynz		| it's  non-zero, skip
	fmoves	#0e1.0,fp0	| Anything**0.0 = 1.0
	jra	powdone	
pynz:
	fmoved	sp@(4),fp1	| fetch X argument
	fjgt	pynzxp		| It's greater than 0, skip
	fjeq	pynzxz		| It's equal to 0, skip
	|
	| Here's where the simple-mindedness comes in.
	| The floating point unit can handle integers
	| much larger than 2^31.
	| Also, integer**integer should be an exact operation, with
	| no exp/log to introduce errors.
	|
	fmovel	fp0,d0		| get int(Y)
	fcmpl	d0,fp0		| does it equal Y?
	fbne	pnan		| no, skip
	movel	d0,sp@-		| push int(Y)
	fnegx	fp1		| get abs(X)
	fmoved	fp0,sp@-	| push Y argument
	fmoved	fp1,sp@-	| push abs(X)
	jbsr	flognd		| form ln(abs(X))
	addl	#8,sp		| pop abs(X)
	fmuld	sp@+,fp0	| form Y*ln(abs(X))
	fmovex	fp0,sp@-	| push Y*ln(abs(X)) (extended)
	jbsr	fetoxx		| form exp(Y*ln(abs(X))) (extended)
	addl	#12,sp		| pop Y*ln(abs(X)) (extended)
	movel	sp@+,d0		| get back int(Y)
	rorl	#1,d0		| even power?
	jcc	powdone		| yes, return value
	fnegx	fp0		| no, return -value
	jra	powdone
pnan:
	fmoves	NANS,fp0	| return NAN
	jra	powdone	
pynzxz:
	ftstx	fp0		| Check Y argument
	fjgt	pypxz		| It's positive, skip
	fmoves	INFS,fp0	| return +infinity
	jra	powdone	
pypxz:
	fmoves	#0e0.0,fp0	| return zero
	jra	powdone	
pynzxp:
	fmoved	fp0,sp@-	| push Y argument
	fmoved	fp1,sp@-	| push X argument
	jbsr	flognd		| form ln(X)
	addl	#8,sp		| pop X argument
	fmuld	sp@+,fp0	| form Y*ln(X)
	fmovex	fp0,sp@-	| push Y*ln(X) (extended)
	jbsr	fetoxx		| form exp(Y*ln(X)) (extended)
	addl	#12,sp		| pop Y*ln(X) (extended)
powdone:
	fmoved	fp0,sp@-	| Archaic NeXT compiler returns doubles in d0/d1
	movel	sp@+,d0
	movel	sp@+,d1
	rts
NANS:	.long 0x7fffffff
INFS:	.long 0x7f800000

	.globl _asinh
_asinh:
	fmoved	sp@(4),fp0	| get argument
	fmulx	fp0,fp0		| X^2
	fadds	#0e1.0,fp0	| X^2+1
	fmovex	fp0,sp@-	| push X^2+1
	jbsr	fsqrtx		| sqrt(X^2+1)
	faddd	sp@(16),fp0	| X+sqrt(X^2+1)
	fmovex	fp0,sp@		| push X+sqrt(X^2+1)
	jbsr	flognx		| ln(X+sqrt(X^2+1))
	addl	#12,sp		| pop X+sqrt(X^2+1)
	fmoved	fp0,sp@-	| Archaic NeXT compiler returns doubles in d0/d1
	movel	sp@+,d0
	movel	sp@+,d1
	rts

	.globl _acosh
_acosh:
	fmoved	sp@(4),fp0	| get argument
	fcmps	#0e1.0,fp0	| greater than 1.0?
	fjgt	1f		| yes, go compute result
	fjne	2f		| less than 1.0, skip
	fmoves	#0e0.0,fp0	| equal to 1.0, just return 0.0
	jra	9f
2:
	fmoves	NANS,fp0	| argument < 1.0, return NAN
	jra	9f
1:
	fmulx	fp0,fp0		| X^2
	fadds	#0e-1.0,fp0	| X^2-1
	fmovex	fp0,sp@-	| push X^2-1
	jbsr	fsqrtx		| sqrt(X^2-1)
	faddd	sp@(16),fp0	| X+sqrt(X^2-1)
	fmovex	fp0,sp@		| push X+sqrt(X^2-1)
	jbsr	flognx		| ln(X+sqrt(X^2-1))
	addl	#12,sp		| pop X+sqrt(X^2-1)
9:
	fmoved	fp0,sp@-	| Archaic NeXT compiler returns doubles in d0/d1
	movel	sp@+,d0
	movel	sp@+,d1
	rts
